import * as ESBuild from 'https://deno.land/x/esbuild@v0.14.45/mod.js';
import * as MIME from "https://deno.land/std@0.180.0/media_types/mod.ts";
import { debounce } from "https://deno.land/std@0.151.0/async/debounce.ts";

console.log(`Serving files from "${Deno.cwd()}"`);

const Transpiled = new Map();
/**
 * checks for (.tsx | .jsx | .ts | .js) extensions
 */
export const Transpileable =(inFilePath:string):boolean=>
{
    let dotIndex = inFilePath.length-4;
    if(inFilePath[dotIndex] !== ".")
    {
        if(inFilePath[++dotIndex] !== ".")
        {
            return false;
        }
    }

    if(inFilePath[dotIndex+2] == "s")
    {
        const first = inFilePath[dotIndex+1];
        return (first == "t" || first == "j");
    }

    return false;
};
export const Transpile =async(inPath:string, inKey:string):Promise<string>=>
{
    const body = await Deno.readTextFile(inPath);
    const transpile = await ESBuild.transform(body, { loader: "tsx", sourcemap: "inline", minify:true });
    Transpiled.set(inKey, transpile.code);
    return transpile.code;
};


let ImportString = ``;
let ImportObject = {};
try
{
    const confDeno = await Deno.readTextFile(Deno.cwd()+"/deno.json");
    const confDenoParsed:{importMap?:string, imports?:string} = JSON.parse(confDeno);
    if(confDenoParsed.importMap)
    {
        try
        {
            const confImports = await Deno.readTextFile(confDenoParsed.importMap);
            try
            {
                ImportObject = JSON.parse(confImports);
                ImportString = confImports;
            }
            catch(e)
            {
                console.log(`importMap "${confDenoParsed.importMap}" contains invalid JSON`);
            }
        }
        catch(e)
        {
            console.log(`importMap "${confDenoParsed.importMap}" cannot be found`);
        }
    }
    else if(confDenoParsed.imports)
    {
        ImportObject = {imports:confDenoParsed.imports};
        ImportString = JSON.stringify(ImportObject);
    }
}
catch(e)
{
    console.log(`deno.json not found`);
}
const Index = `
<!doctype html>
<html>
    <head>
        <meta name="viewport" content="width=device-width, initial-scale=1.0">
    </head>
    <body>
        <div id="app">Loading</div>
        <script type="importmap">${ImportString}</script>
        <script async src="https://ga.jspm.io/npm:es-module-shims@1.5.1/dist/es-module-shims.js" crossorigin="anonymous"></script>
        <script type="module">
            import * as TW   from "https://esm.sh/@twind/core@1.0.1";
            import TWPreTail from "https://esm.sh/@twind/preset-tailwind@1.0.1";
            import TWPreAuto from "https://esm.sh/@twind/preset-autoprefix@1.0.1";
            
            const Configure = {presets: [TWPreTail(), TWPreAuto()]};
            const ShadowDOM = document.querySelector("#app").attachShadow({mode: "open"});
            const ShadowDiv = document.createElement("div");
            const ShadowCSS = document.createElement("style");
            ShadowDOM.append(ShadowCSS);
            ShadowDOM.append(ShadowDiv);

            import App from "./app.tsx";
            App(ShadowDiv, Configure);
            TW.observe(TW.twind(Configure, TW.cssom(ShadowCSS)), ShadowDiv);
        </script>
    </body>
</html>
`;

Deno.serve({ port: 3000 }, async(_req:Request) =>
{
    const url:URL = new URL(_req.url);
    const fsPath = Deno.cwd()+url.pathname;

    console.log(`Request for "${url.pathname}"...`);

    try
    {
        // serve index by default
        let type = `text/html`;
        let body:BodyInit = Index;

        // serve .tsx .jsx .ts .js
        if(Transpileable(url.pathname))
        {
            body = Transpiled.get(url.pathname);
            type = `application/javascript`;
            if(!body)
            {
                body = await Transpile(fsPath, url.pathname);
            }
        }
        // serve static media
        else if( url.pathname.endsWith("/") === false)
        {
            body = await Deno.readFile(fsPath);
            type = MIME.typeByExtension(url.pathname.substring(url.pathname.lastIndexOf("."))) || "text/html";
        }

        return new Response(body, {headers:{"content-type":type as string}});
    }
    catch(error)
    {
        console.log(`  ...404`);
        return new Response(error, {status:404});
    }
});


const Sockets:Set<WebSocket> = new Set();
const SocketsBroadcast =(inData:string)=>{ for (const socket of Sockets){ socket.send(inData); } }
Deno.serve({ port: 3001 }, (_req:Request) =>
{
    if(_req.headers.get("upgrade") == "websocket")
    {
        try
        {
          const { response, socket } = Deno.upgradeWebSocket(_req);
          socket.onopen = () =>
          {
              Sockets.add(socket);
              console.log("Overwatch: Socket created");
          };
          socket.onclose = () =>
          {
              Sockets.delete(socket);
              console.log("Overwatch: Socket deleted");
          };
          socket.onmessage = (e) => {};
          socket.onerror = (e) => console.log("Overwatch: Socket errored:", e);
          return response;
        }
        catch(e)
        {
            //
        }
    }
    return new Response("Overwatch: Not a websocket request.");
});
const FilesChanged:Map<string, string> = new Map();
const ProcessFiles =debounce(async()=>
{
    console.log("Files changed...")
    for await (const [file, action] of FilesChanged)
    {
        const pathname = file.substring(Deno.cwd().length).replaceAll("\\", "/");
        console.log(pathname, action);
        if(Transpileable(pathname))
        {
            if(action !== "remove")
            {
                await Transpile(file, pathname);
                console.log(`  ...cached "${pathname}"`);
                SocketsBroadcast(pathname);
            }
        }
    }
    FilesChanged.clear();
}, 1000);
for await (const event of Deno.watchFs(Deno.cwd()))
{
    event.paths.forEach( path => FilesChanged.set(path, event.kind) );
    ProcessFiles();
}